# coding: utf-8
import pandas as pd
from tqdm import tqdm
from typing import List, Callable, Tuple


class FactorsResearcher:
    """因子研究员"""

    def __init__(self, factors, n_list=None):
        self.factors = factors
        self.factors_df = pd.DataFrame(factors)
        self.factors_df['base'] = 1
        self.performance = None
        if not n_list:
            self.n_list = (15, 30, 60, 120, 240, 480, 960)
        else:
            self.n_list = n_list
        self.__add_nbar()
        self.base = self.factor_performance(['base']).iloc[0].to_dict()

    def __add_nbar(self):
        """计算 nbar 相关特征"""
        df = self.factors_df
        n_list = self.n_list
        for bar_number in n_list:
            n_col_name = 'n' + str(bar_number) + 'b'
            df[n_col_name] = (df['close'].shift(-bar_number) / df['close'] - 1) * 10000
        self.factors_df = df

    def factor_performance(self, cols: List):
        """查看因子组合的表现"""
        df = self.factors_df
        df = df.dropna()
        n_list = self.n_list

        results = []
        for values, dfg in df.groupby(cols):
            res = {"factor_name": "&".join(cols)}
            if isinstance(values, tuple):
                for i, v in enumerate(values, 1):
                    res['factor_value{}'.format(i)] = v
            else:
                res['factor_value'] = values

            res["count"] = len(dfg)

            for n in n_list:
                nb_col = 'n{}b'.format(n)
                dn = dfg[nb_col]
                dn1 = dfg[dfg[nb_col] >= 0]
                dn2 = dfg[dfg[nb_col] < 0]
                res['n{}b均收益'.format(n)] = round(dn.mean(), 2)
                if dn2.empty:
                    res['n{}b盈亏比'.format(n)] = round(dn1[nb_col].mean(), 2)
                elif dn1.empty:
                    res['n{}b盈亏比'.format(n)] = round(dn2[nb_col].mean(), 2)
                else:
                    res['n{}b盈亏比'.format(n)] = round(abs(dn1[nb_col].mean() / (dn2[nb_col].mean() + 1)), 2)
                res['n{}b胜率'.format(n)] = round(len(dn1) / len(dn), 2)

            results.append(res)
        df_res = pd.DataFrame(results)
        df_res['pct'] = df_res['count'] / len(df)
        return df_res

    def single_factor_analyze(self, cols: [list, tuple]):
        """单因子分析"""
        data = []
        for col in tqdm(cols, desc="single_factor_analyze"):
            data.append(self.factor_performance([col]))
        df = pd.concat(data)
        return df

    def pair_factor_analyze(self, pairs:  List[Tuple[str, str]]):
        """两因子分析"""
        data = []
        for pair in tqdm(pairs, desc="pair_factor_analyze"):
            data.append(self.factor_performance(list(pair)))
        df = pd.concat(data)
        return df

    def triple_factor_analyze(self, triples:  List[Tuple[str, str, str]]):
        """三因子分析"""
        data = []
        for triple in tqdm(triples, desc="triple_factor_analyze"):
            data.append(self.factor_performance(list(triple)))
        df = pd.concat(data)
        return df

    def add_temp_factors(self, functions: List[Callable]) -> None:
        """新增因子，用于快速研究

        :param functions:
        :return: None
        """
        for factor in functions:
            name = factor.__doc__
            for row in tqdm(self.factors, desc="add {}".format(name)):
                row[name] = factor(row)





# coding: utf-8
import pandas as pd
from datetime import datetime
from .factors import KlineGeneratorBy1Min, CzscFactors
from .data.jq import get_kline, get_kline_period
from .data import freq_map, freq_inv
from .enum import Signals, Factors

class CzscTrader:
    """缠中说禅股票 选股/择时"""
    def __init__(self, symbol, max_count=1000, end_date=None):
        """
        :param symbol:
        """
        self.symbol = symbol
        if end_date:
            self.end_date = pd.to_datetime(end_date)
        else:
            self.end_date = datetime.now()
        self.max_count = max_count
        kg = KlineGeneratorBy1Min(max_count=max_count*2, freqs=['1分钟', '5分钟', '15分钟', '30分钟', '60分钟', '日线'])
        for freq in kg.freqs:
            bars = get_kline(symbol, end_date=self.end_date, freq=freq_inv[freq], count=max_count)
            kg.init_kline(freq, bars)
        kf = CzscFactors(kg)
        self.kf = kf
        self.s = kf.s
        self.freqs = kg.freqs

    def __repr__(self):
        return "<CzscTrader of {} @ {}>".format(self.symbol, self.kf.end_dt)

    def run_selector(self):
        """输出日线笔因子"""
        s = self.s
        factors_d = [x.value for x in Factors.__members__.values() if x.name[:2] == 'C6']
        if s['日线笔因子'] in factors_d:
            return s['日线笔因子']
        return "other"

    def run_history(self):
        """查看第N-3笔的历史形态"""
        s = self.s
        nine_values = [x.value for x in Signals.__members__.values() if x.name[:3] in ["X9L", "X9S"]]
        seven_values = [x.value for x in Signals.__members__.values() if x.name[:3] in ["X7L", "X7S"]]
        five_values = [x.value for x in Signals.__members__.values() if x.name[:3] in ["X5L", "X5S"]]

        for freq in ["30分钟", "日线"]:
            if s['{}_倒4的九笔形态'.format(freq)] in nine_values:
                return "{}_倒4的九笔形态_{}".format(freq, s['{}_倒4的九笔形态'.format(freq)])

            if s['{}_倒4的七笔形态'.format(freq)] in seven_values:
                return "{}_倒4的七笔形态_{}".format(freq, s['{}_倒4的七笔形态'.format(freq)])

            if s['{}_倒4的五笔形态'.format(freq)] in five_values:
                return "{}_倒4的五笔形态_{}".format(freq, s['{}_倒4的五笔形态'.format(freq)])
        return "other"

    def take_snapshot(self, file_html, width="1400px", height="680px"):
        self.kf.take_snapshot(file_html, width, height)

    def open_in_browser(self, width="1400px", height="580px"):
        self.kf.open_in_browser(width, height)

    def update_factors(self):
        """更新K线数据到最新状态"""
        bars = get_kline_period(symbol=self.symbol, start_date=self.kf.end_dt, end_date=datetime.now(), freq="1min")
        if not bars:
            return
        for bar in bars:
            self.kf.update_factors([bar])
        self.s = self.kf.s





# coding: utf-8
from collections import OrderedDict
from pyecharts.charts import Tab
from pyecharts.components import Table
from pyecharts.options import ComponentTitleOpts
import os
import webbrowser
from typing import List
from .analyze import CZSC, get_sub_bis, get_sub_span
from .utils.kline_generator import KlineGeneratorBy1Min, KlineGeneratorByTick
from .objects import RawBar
from .enum import Factors, Signals, Direction


def check_triple_level(c1: CZSC, c2: CZSC, c3: CZSC):
    """三级别联立笔因子计算

    c1, c2, c3 可能的组合
    1）日线、60分钟、15分钟
    2）日线、30分钟、5分钟
    3）日线、15分钟、1分钟
    4）60分钟、15分钟、1分钟
    5）60分钟、5分钟、1分钟

    :param c1: 本级别 CZSC 对象
    :param c2: 次级别 CZSC 对象
    :param c3: 小级别 CZSC 对象
    :return:
    """
    c_map = {"日线": "C6", "60分钟": "C5", "30分钟": "C4", "15分钟": "C3", "5分钟": "C2", "1分钟": "C1"}
    factor_key_base = "{}{}".format(c_map[c1.freq], c_map[c2.freq])

    v = Factors.Other.value
    c2_h9 = max([x.high for x in c2.bi_list[-9:]])
    c2_l9 = max([x.high for x in c2.bi_list[-9:]])

    if c1.bi_list[-1].direction == Direction.Down and len(c1.bars_ubi) <= 7:
        factor_l1 = factor_key_base + "L1"  # L1 - 向下笔转折右侧
        if c2.signals['倒1的五笔形态'] == Signals.X5LB0.value:
            if c2_h9 == c2.bi_list[-1].high:
                v = Factors['{}A1'.format(factor_l1)].value
            elif "顶背弛" in c1.signals['倒2的五笔形态'] \
                    and "顶背弛" in c1.signals['倒2的七笔形态'] \
                    and "顶背弛" in c1.signals['倒2的九笔形态']:
                v = Factors['{}A2'.format(factor_l1)].value
            elif "底背弛" in c1.signals['倒1的五笔形态'] \
                    and "底背弛" in c1.signals['倒1的七笔形态'] \
                    and "底背弛" in c1.signals['倒1的九笔形态']:
                v = Factors['{}A3'.format(factor_l1)].value
            else:
                v = Factors['{}A0'.format(factor_l1)].value

        if c1.bi_list[-1].fx_b.power == "强" and c2.signals['倒1的七笔形态'] == Signals.X7LE0.value:
            v = Factors['{}B0'.format(factor_l1)].value

        if v != Factors.Other.value:
            return v

    if c1.bi_list[-1].direction == Direction.Up and len(c1.bars_ubi) > 7:
        factor_l2 = factor_key_base + "L2"
        if "底背弛" in c2.signals['倒1的五笔形态']:
            v = Factors['{}A0'.format(factor_l2)].value

        if "底背弛" in c2.signals['倒1的七笔形态']:
            v = Factors['{}B0'.format(factor_l2)].value

        if "底背弛" in c2.signals['倒1的九笔形态']:
            v = Factors['{}C0'.format(factor_l2)].value
            if c2.signals['倒1的九笔形态'] == Signals.X9LA0.value:
                v = Factors['{}C1'.format(factor_l2)].value

        if v != Factors.Other.value:
            return v

    if c1.bi_list[-1].direction == Direction.Up and len(c1.bars_ubi) <= 7 \
            and min([x.low for x in c1.bars_ubi]) > c1.bi_list[-1].low:
        if c3.signals['倒1的五笔形态'] in [Signals.X5LF0.value, Signals.X5LB0.value]:
            factor_l3 = factor_key_base + "L3"
            if c2.signals['倒1的七笔形态'] == Signals.X7LE0.value:
                v = Factors['{}A0'.format(factor_l3)].value
        else:
            factor_l4 = factor_key_base + "L4"
            if "底背弛" in c2.signals['倒1的七笔形态']:
                v = Factors['{}A0'.format(factor_l4)].value

        if v != Factors.Other.value:
            return v

    if c1.bi_list[-1].direction == Direction.Up and len(c1.bars_ubi) <= 7:
        factor_s1 = factor_key_base + "S1"
        if c2.signals['倒1的五笔形态'] == Signals.X5SB0.value:
            if c2_l9 == c2.bi_list[-1].low:
                v = Factors['{}A1'.format(factor_s1)].value
            elif "底背弛" in c1.signals['倒2的五笔形态'] \
                    and "底背弛" in c1.signals['倒2的七笔形态'] \
                    and "底背弛" in c1.signals['倒2的九笔形态']:
                v = Factors['{}A2'.format(factor_s1)].value
            elif "顶背弛" in c1.signals['倒1的五笔形态'] \
                    and "顶背弛" in c1.signals['倒1的七笔形态'] \
                    and "顶背弛" in c1.signals['倒1的九笔形态']:
                v = Factors['{}A3'.format(factor_s1)].value
            else:
                v = Factors['{}A0'.format(factor_s1)].value

        if c2.signals['倒1的五笔形态'] == Signals.X5SA0.value:
            v = Factors['{}B0'.format(factor_s1)].value

        if v != Factors.Other.value:
            return v

    if c1.bi_list[-1].direction == Direction.Down and len(c1.bars_ubi) > 7:
        factor_s2 = factor_key_base + "S2"
        if "顶背驰" in c2.signals['倒1的七笔形态']:
            v = Factors['{}A0'.format(factor_s2)].value

    if c1.bi_list[-1].direction == Direction.Down and len(c1.bars_ubi) <= 7 \
            and max([x.high for x in c1.bars_ubi]) < c1.bi_list[-1].high:
        if c3.signals['倒1的五笔形态'] in [Signals.X5SF0.value, Signals.X5SB0.value]:
            factor_s3 = factor_key_base + "S3"
            if c2.signals['倒1的五笔形态'] == Signals.X5SF0.value:
                v = Factors['{}A0'.format(factor_s3)].value
        else:
            factor_s4 = factor_key_base + "S4"
            if "顶背弛" in c3.signals['倒1的五笔形态']:
                v = Factors['{}A0'.format(factor_s4)].value
    return v

class CzscFactors:
    """缠中说禅技术分析理论之多级别联立因子"""
    def __init__(self, kg: [KlineGeneratorByTick, KlineGeneratorBy1Min]):
        """

        :param kg: 基于tick或1分钟的K线合成器
        """
        self.kg = kg
        self.freqs = kg.freqs
        klines = self.kg.get_klines({k: 3000 for k in self.freqs})
        self.kas = {k: CZSC(klines[k], freq=k, max_bi_count=20) for k in klines.keys()}
        self.symbol = self.kas["1分钟"].symbol
        self.end_dt = self.kas["1分钟"].bars_raw[-1].dt
        self.latest_price = self.kas["1分钟"].bars_raw[-1].close
        self.s = self._calculate_signals()
        self.s = self._calculate_factors_d()
        self.s = self._calculate_factors_f60()
        self.cache = OrderedDict()

    def take_snapshot(self, file_html=None, width="1400px", height="580px"):
        """获取快照

        :param file_html: str
            交易快照保存的 html 文件名
        :param width: str
            图表宽度
        :param height: str
            图表高度
        :return:
        """
        tab = Tab(page_title="{}@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            chart = self.kas[freq].to_echarts(width, height)
            tab.add(chart, freq)

        t1 = Table()
        t1.add(["名称", "数据"], [[k, v] for k, v in self.s.items() if "_" in k and "~" in str(v)])
        t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
        tab.add(t1, "信号表")

        t2 = Table()
        t2.add(["名称", "数据"], [[k, v] for k, v in self.s.items() if "_" not in k and v != "其他"])
        t2.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅因子表", subtitle=""))
        tab.add(t2, "因子表")

        if file_html:
            tab.render(file_html)
        else:
            return tab

    def open_in_browser(self, width="1400px", height="580px"):
        """直接在浏览器中打开分析结果"""
        home_path = os.path.expanduser("~")
        file_html = os.path.join(home_path, "temp_czsc_factors.html")
        self.take_snapshot(file_html, width, height)
        webbrowser.open(file_html)

    def _calculate_signals(self):
        """计算信号"""
        s = OrderedDict()
        for freq, ks in self.kas.items():
            # s.update(ks.signals)
            s.update({"{}_{}".format(ks.freq, k) if k not in ['symbol', 'dt', 'close'] else k: v
                      for k, v in ks.signals.items()})

        s.update(self.kas['1分钟'].bars_raw[-1].__dict__)
        return s

    def _calculate_factors_d(self):
        s = self.s
        s.update({"日线笔因子": Factors.Other.value})
        c1: CZSC = self.kas['1分钟']
        c2: CZSC = self.kas['5分钟']
        c3: CZSC = self.kas['15分钟']
        c4: CZSC = self.kas['30分钟']
        c5: CZSC = self.kas['60分钟']
        c6: CZSC = self.kas['日线']

        if not c6.bi_list:
            print("{} 日线笔数量为 0".format(self.symbol))
            return s

        # 感知日线笔的真实级别，确定 c1, c2, c3
        if len(c6.bars_ubi) > 9:
            direction = Direction.Down if c6.bi_list[-1].direction == Direction.Up else Direction.Up
            bis_c6_c4 = get_sub_span(c4.bi_list[-15:], start_dt=c6.bars_ubi[1].dt,
                                     end_dt=c6.bars_ubi[-1].dt, direction=direction)
        else:
            bis_c6_c4 = get_sub_bis(c4.bi_list[-15:], c6.bi_list[-1])

        if 9 >= len(bis_c6_c4) >= 5:
            c1_, c2_, c3_ = c6, c4, c2
        elif len(bis_c6_c4) > 9:
            c1_, c2_, c3_ = c6, c5, c3
        elif len(bis_c6_c4) < 5:
            c1_, c2_, c3_ = c6, c3, c1
        else:
            raise ValueError

        s.update({"日线笔因子": check_triple_level(c1_, c2_, c3_)})
        return s

    def _calculate_factors_f60(self):
        """计算因子"""
        s = self.s
        s.update({"60分钟笔因子": Factors.Other.value})
        c1: CZSC = self.kas['1分钟']
        c2: CZSC = self.kas['5分钟']
        c3: CZSC = self.kas['15分钟']
        c5: CZSC = self.kas['60分钟']

        if not c5.bi_list:
            print("{} 60分钟笔数量为 0".format(self.symbol))
            return s

        # 感知60分钟笔的真实级别，确定 c1, c2, c3
        if len(c5.bars_ubi) > 9:
            direction = Direction.Down if c5.bi_list[-1].direction == Direction.Up else Direction.Up
            bis_c5_c4 = get_sub_span(c3.bi_list[-15:], start_dt=c5.bars_ubi[1].dt,
                                     end_dt=c5.bars_ubi[-1].dt, direction=direction)
        else:
            bis_c5_c4 = get_sub_bis(c3.bi_list[-15:], c5.bi_list[-1])

        if len(bis_c5_c4) >= 5:
            c1_, c2_, c3_ = c5, c3, c1
        else:
            c1_, c2_, c3_ = c5, c2, c1

        s.update({"60分钟笔因子": check_triple_level(c1_, c2_, c3_)})
        return s

    def update_factors(self, data: List[RawBar]):
        """更新多级别联立因子"""
        for row in data:
            self.kg.update(row)

        klines_one = self.kg.get_klines({k: 1 for k in self.freqs})

        for freq, klines_ in klines_one.items():
            k = klines_[-1]
            self.kas[freq].update(k)

        self.symbol = self.kas["1分钟"].symbol
        self.end_dt = self.kas["1分钟"].bars_raw[-1].dt
        self.latest_price = self.kas["1分钟"].bars_raw[-1].close
        self.s = self._calculate_signals()
        self.s = self._calculate_factors_d()
        self.s = self._calculate_factors_f60()

